#!/usr/bin/env python3

# This script is written in Python 3. You need a recent version.
# Windows users can install from https://www.python.org/ and
# macOS and Linux users most likely have it already installed.

# macOS and Linux: Run "python3 wadutil.py --help" in a Terminal
# Windows: Run "py wadutil.py --help" in cmd.exe or Windows Terminal

# The lzss compression is not the fastest (but might very well be
# the slowest implementation there is...). You can get a performance
# boost by using PyPy3 (https://www.pypy.org/) which has a JIT. Of
# course, there's lots of potential to optimize the algorithm; this
# reference implementation is mostly about documenting how it works.

# This script comes with a "filenames.txt" list of file names, which
# will be used to build a rainbow table for reversing the file names.
# Not all filenames are known (yet). If you find out a filename for
# a hash in the official game and DLC files, let me (m@thp.io) know.

# So without further ado, here it is:

"""
   _ _ _ _                 _      _____ _____ _____
  | | | |_|___ ___ ___ _ _| |_   |  _  |   __|  _  |
  | | | | | . | -_| . | | |  _|  |   __|__   |   __|
  |_____|_|  _|___|___|___|_|    |__|  |_____|__|
       _ _|_|_____ ____     _____ _   _ _ _ _
      | | | |  _  |    \   |  |  | |_|_| |_| |_ _ _
      | | | |     |  |  |  |  |  |  _| | | |  _| | |
      |_____|__|__|____/   |_____|_| |_|_|_|_| |_  |
                                               |___|
  https://thp.io/2021/wadutil/              (v1.0.0)
  2021-09-17; Thomas Perl <m@thp.io>

"""

#
# WipEout PSP WAD Utility -- List, extract and create WAD files
# Copyright 2019, 2020, 2021 Thomas Perl <m@thp.io>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

# Parts based on the Font Extractor 2019-10-10 thp
# Initial tool version: wadextract.py 2020-10-20 thp
# Initial WAD re-packing prototype: 2020-11-19 thp
# LZSS Decompressor (wadunpack.py): 2021-01-03 thp
# Bruteforce (slow) LZSS Compressor: 2021-01-09 thp
# Cleanup, documentation and release: 2021-09-17 thp

# Documentation on the file format:
# https://thp.io/2020/pure-wad.html

# Documentation on filename hashing:
# https://thp.io/2020/wad-hash.html


import zlib
import re
import struct
import os
import argparse


HERE = os.path.abspath(os.path.dirname(os.path.normpath(os.path.realpath(__file__))) or '.')


def filename_hash(filename):
    filename = ''.join('/' if ch == '\\' else ch.lower() for ch in filename)
    m = re.match(r'^data/unknown/([0-9a-f]{8})\.raw$', filename)
    return int(m.group(1), 16) if m is not None else zlib.crc32(filename.encode(), 0xFFFFFFFF)


class DecompressionContext:
    @classmethod
    def unpack_compressed(cls, compressed_data, uncompressed_size):
        return bytes(cls(compressed_data).unpack(uncompressed_size))

    def __init__(self, input_buffer):
        self.input_buffer = input_buffer
        self.lookback_buffer_write_index = 1
        self.lookback_buffer = bytearray(0x2000)

    def bit_stream(self):
        for ch in self.input_buffer:
            for i in range(8):
                yield 1 if (ch & (1 << (7 - i))) else 0

    def unpack(self, length):
        bit_stream = self.bit_stream()

        def read_bits(stream, n):
            return sum((1<<(n-1-i)) * next(stream) for i in range(n))

        while length > 0:
            if next(bit_stream):
                # if bit 1 is set, it's a verbatim byte
                tmp = read_bits(bit_stream, 8)
                yield tmp
                length -= 1

                self.lookback_buffer[self.lookback_buffer_write_index] = tmp
                self.lookback_buffer_write_index = (self.lookback_buffer_write_index + 1) & 0x1fff
            else:
                # if bit 1 is not set, it's a copy from previous contents:
                # 13 bits offset
                # 4 bits repetition count (- 3) -- up to 18 bytes repeated
                copy_from_lookback_offset = read_bits(bit_stream, 13)
                repetition_count = 3 + read_bits(bit_stream, 4)
                for i in range(repetition_count):
                    tmp = self.lookback_buffer[(copy_from_lookback_offset + i) & 0x1fff]
                    yield tmp
                    length -= 1

                    self.lookback_buffer[self.lookback_buffer_write_index] = tmp
                    self.lookback_buffer_write_index = (self.lookback_buffer_write_index + 1) & 0x1fff


def decompress_lzss(data, uncompressed_size):
    return DecompressionContext.unpack_compressed(data, uncompressed_size)


def compress_lzss(data):
    BITS_VERBATIM = 8

    LOOKBACK_LENGTH = 0x2000

    BITS_LENGTH = 4
    REPETITION_ADD = 3
    MAX_REPETITION = REPETITION_ADD + ((1 << BITS_LENGTH) - 1)

    BITS_OFFSET = 13
    MAX_OFFSET = (1 << BITS_OFFSET) - 1

    def bits(n, v):
        for i in range(n):
            yield 1 if v&(1<<(n-i-1)) else 0

    def lzss_encode(data):
        last_pct = -1
        pos = 0
        while pos < len(data):
            best_len = 0
            best_off = 0

            nmax = min(pos, MAX_OFFSET)
            for o in range(1, nmax):
                l = 0
                while l < MAX_REPETITION and l < len(data)-pos and data[pos+l] == data[pos-o+l]:
                    l += 1

                if l > best_len:
                    best_off = o
                    best_len = l

                if best_len == MAX_REPETITION:
                    # TODO: Early exit possible, but is this how the WOP algo works?
                    break

            if best_len >= REPETITION_ADD:
                offset = (pos - best_off + 1) & (LOOKBACK_LENGTH-1)
                yield from bits(1, 0) # mode bit

                yield from bits(BITS_OFFSET, offset)
                yield from bits(BITS_LENGTH, best_len - REPETITION_ADD)

                pos += best_len
            else:
                yield from bits(1, 1) # mode bit

                yield from bits(BITS_VERBATIM, data[pos])

                pos += 1

    def bitstream_to_bytes(bitstream):
        nbits = 0
        cur = 0
        for bit in bitstream:
            cur |= (bit << (7-nbits))
            nbits += 1
            if nbits == 8:
                yield cur
                nbits = 0
                cur = 0

        if nbits != 0:
            yield cur

    return bytes(bitstream_to_bytes(lzss_encode(data)))


MODE_STORED = 'stored'
MODE_LZSS = 'lzss'
MODE_ZLIB = 'zlib'


def iterate_wad(wad, need_data):
    hdrsz = struct.calcsize('<II')
    version, nfiles = struct.unpack('<II', wad[:hdrsz])

    assert version == 1, 'Not a WAD file (encrypted Pure DLC?)'

    print(f'Number of files: {nfiles}')

    entrysz = struct.calcsize('<IIII')

    expected_offset = hdrsz + entrysz * nfiles

    off = hdrsz
    for i in range(nfiles):
        name_hash, offset, uncompressed_length, compressed_length = struct.unpack('<IIII', wad[off:off+entrysz])
        off += entrysz

        if offset != expected_offset:
            assert all(bytevalue == 0 for bytevalue in wad[expected_offset:offset])

        expected_offset = offset + compressed_length

        d = wad[offset:offset+compressed_length]

        if uncompressed_length == compressed_length:
            mode = MODE_STORED
        elif uncompressed_length & 0x80000000 == 0:
            mode = MODE_LZSS
            if need_data:
                d = decompress_lzss(d, uncompressed_length)
                assert len(d) == uncompressed_length
        else:
            # Pure EU DLC Packs support zlib compression
            mode = MODE_ZLIB
            uncompressed_length &= ~0x80000000
            if need_data:
                d = zlib.decompress(d)
                assert len(d) == uncompressed_length

        assert offset + compressed_length <= len(wad)

        if not need_data:
            d = None

        yield (offset, d, name_hash, mode, compressed_length, uncompressed_length)


class FilenameResolver:
    def __init__(self, filenames_txt):
        self.filenames = {filename_hash(filename): filename
                          for filename in open(filenames_txt).read().strip().split('\n')}

    def lookup(self, name_hash):
        return self.filenames.get(name_hash, f'data/unknown/{name_hash:08x}.raw')


def main():
    # seems to be 64 in built-in files, 128 in plugin wad files
    # It might very well be that this doesn't matter at all, and
    # tightly-packed files are fine.
    target_alignment = 0x80

    parser = argparse.ArgumentParser(description='List, extract and create WAD files for PSP WipEout games.')
    parser.add_argument('--extract', '-x', action='store_true', help='Extract a WAD file (to the current directory)')
    parser.add_argument('--create', '-c', metavar='output.wad', type=str, help='Create a WAD file from a folder')
    parser.add_argument('--zlib-compress', action='store_true', help='Enable zlib compression (EUR only, for --create)')
    parser.add_argument('--align', type=int, default=target_alignment, help=f'Align start offsets (default: {target_alignment} bytes)')
    parser.add_argument('--no-compress', action='store_true', help='Do not compress anything (for --create)')
    parser.add_argument('source', type=str, help='Input name (either WAD filename or folder name)')

    args = parser.parse_args()

    target_alignment = args.align

    if args.zlib_compress and not args.create:
        print('ERROR: --zlib-compress only makes sense with --create')
        raise SystemExit(1)

    if args.no_compress and not args.create:
        print('ERROR: --no-compress only makes sense with --create')
        raise SystemExit(1)

    if args.create is not None and args.extract:
        print('ERROR: Cannot create and extract at the same time')
        raise SystemExit(1)

    # Create a kind of rainbow table to "reverse" hashed file names to their original value
    resolver = FilenameResolver(os.path.join(HERE, 'filenames.txt'))

    if args.create:
        source_folder = args.source
        output_wad = args.create

        filenames = []
        original_filelist = None

        for dirpath, dirnames, fns in os.walk(source_folder):
            dest_folder = dirpath[len(source_folder):].lstrip('/')
            for filename in sorted(fns):
                fn = os.path.join(source_folder, filename)
                if filename == '__filelist__':
                    original_filelist = [re.match(r'^(\d+) (\d+) (stored|lzss|zlib) (.*)$', line).groups()
                                for line in open(fn).read().splitlines()]
                    continue
                filenames.append(os.path.join(dest_folder, filename))

        def align_to(value, alignment):
            while alignment != 0 and value % alignment != 0:
                value += 1
            return value

        if original_filelist is not None:
            original_filenames = [filename for _, _, _, filename in original_filelist]
            original_filelist = {filename: (int(offset), int(compressed_length), mode)
                                 for offset, compressed_length, mode, filename in original_filelist}

            print('Re-using __filelist__ for alignment/offset and compression')
            if set(filenames) != set(original_filenames):
                print('\nCannot continue, files were added or removed.')
                added_files = set(filenames).difference(set(original_filenames))
                if added_files:
                    print('\nAdded files not appearing in __filelist__:')
                    for added_file in added_files:
                        print(f'\t{added_file}')

                removed_files = set(original_filenames).difference(set(filenames))
                if removed_files:
                    print('\nRemoved files appearing in __filelist__:')
                    for removed_file in removed_files:
                        print(f'\t{removed_file}')

                print('\nEither delete/restore these files or remove __filelist__ to repack.')
                raise SystemExit(1)
            filenames = original_filenames

        offset = align_to(struct.calcsize('<II') + len(filenames) * struct.calcsize('<IIII'), target_alignment)

        entries = []

        header_printed = False
        for filename in filenames:
            data = open(os.path.join(source_folder, filename), 'rb').read()

            # TODO: It seems like .at3 files (music.at3) are never compressed,
            # but this might be because it's already compressed audio data.
            # TODO: Alignment for at3 files should be at least 128 bytes(?)
            can_compress = len(data) and not filename.endswith('.at3') and not args.no_compress

            # The root files seem to be never compressed using zlib, even in EUR
            can_zlib = filename not in ('definition.xml', 'stringtable.xml')

            if original_filelist is not None:
                original_offset, original_compressed_length, original_mode = original_filelist[filename]
                if original_offset > offset:
                    print(f'Moving {filename} by {original_offset-offset} bytes '
                          f'from 0x{offset:08x} to 0x{original_offset:08x}')
                    offset = original_offset
                if original_mode == 'stored':
                    can_compress = False

            flags = 0x00000000
            compressed = data
            mode = MODE_STORED

            if can_compress:
                if can_zlib and args.zlib_compress:
                    compressed = zlib.compress(data, level=9)
                    flags |= 0x80000000
                    mode = MODE_ZLIB
                else:
                    lzss_data = compress_lzss(data)
                    if len(lzss_data) < len(data):
                        compressed = lzss_data
                        mode = MODE_LZSS

            entries.append({
                'filename': filename,
                'data': compressed,
                'offset': offset,
                'uncompressed_size': len(data) | flags,
                'compressed_size': len(compressed),
            })

            if not header_printed:
                print()
                print('OFFSET    NAMEHASH  METHOD  ORIG-SIZE -> COMP-SIZE  RATIO  FILENAME')
                print('========  ========  ======  =========    =========  =====  ========')
                header_printed = True

            ratio = f'{100 * (len(compressed) / len(data)):3.0f}%' if (len(data) and mode != MODE_STORED) else ' ---'
            compinfo = f'{len(data):8d}     {len(compressed):8d}'
            print(f'{offset:08x}  {filename_hash(filename):08x}  {mode:6s}   {compinfo}   {ratio}  {filename}')

            offset += align_to(len(compressed), target_alignment)

        print("""
    NOTE ABOUT INFORMATION LOSS:

    WAD files only store a 32-bit hash of the filename.
    For names not appearing in filenames.txt, the original
    file name cannot be recovered from the WAD file itself.
        """)

        with open(output_wad, 'wb') as fp:
            # version, nfiles
            fp.write(struct.pack('<II', 1, len(entries)))

            for entry in entries:
                fp.write(struct.pack('<IIII', filename_hash(entry['filename']),
                                     entry['offset'], entry['uncompressed_size'], entry['compressed_size']))

            for entry in entries:
                while fp.tell() < entry['offset']:
                    fp.write(b'\0')
                assert entry['offset'] == fp.tell()
                fp.write(entry['data'])

            # trailing padding (might not be needed?)
            while target_alignment != 0 and fp.tell() % target_alignment != 0:
                fp.write(b'\0')
    else:
        filenames = []
        filelist = []

        header_printed = False

        wad = iterate_wad(open(args.source, 'rb').read(), args.extract)
        for offset, data, name_hash, mode, compressed_length, uncompressed_length in wad:
            filename = resolver.lookup(name_hash)
            filenames.append(filename)

            if not header_printed:
                print()
                print('OFFSET    NAMEHASH  METHOD  ORIG-SIZE -> COMP-SIZE  RATIO  FILENAME')
                print('========  ========  ======  =========    =========  =====  ========')
                header_printed = True
            ratio = f'{100 * (compressed_length / uncompressed_length):3.0f}%' if (uncompressed_length and mode != MODE_STORED) else ' ---'
            compinfo = f'{uncompressed_length:8d}     {compressed_length:8d}'
            print(f'{offset:08x}  {name_hash:08x}  {mode:6s}   {compinfo}   {ratio}  {filename}')

            if args.extract:
                filelist.append((offset, compressed_length, mode, filename))

                os.makedirs(os.path.dirname(filename) or '.', exist_ok=True)
                with open(filename, 'wb') as fp:
                    fp.write(data)

        if args.extract:
            with open('__filelist__', 'w') as filelist_txt:
                for offset, compressed_length, mode, filename in filelist:
                    print(f'{offset} {compressed_length} {mode} {filename}', file=filelist_txt)

        unknown_filenames = len([filename for filename in filenames if filename.startswith('data/unknown/')])
        known_filenames = len(filenames) - unknown_filenames
        print(f'\nFilenames: {known_filenames} known, {unknown_filenames} unknown '
              f'-> {100*known_filenames/(known_filenames+unknown_filenames):.2f} % known')

if __name__ == '__main__':
    print(__doc__)
    main()
